# -*- coding: utf-8 -*-
#
###
# Application: wah!cade
# File:        wahcade.py
# Description: Main Window
# Copyright (c) 2007   Andy Balcombe <http://www.anti-particle.com>
###
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
import os
import sys
import glob
import random
import time
import gc
from operator import itemgetter
from subprocess import Popen

#gtk
import pygtk
if sys.platform != 'win32':
    pygtk.require('2.0')
import gtk
import gtk.glade
import gobject
import pango

#project modules
from constants import *
from scrolled_list import ScrollList
from key_consts import mamewah_keys
from wc_common import WahCade
from win_options import WinOptions
from win_message import WinMessage
from win_scrsaver import WinScrSaver
from win_history import WinHistory
from win_cpviewer import get_viewer_window
import filters
from mamewah_ini import MameWahIni
gst_video_imported = False
if sys.platform != 'win32':
    try:
        import gst_video
        gst_video_imported = True
    except:
        pass
_ = gettext.gettext

class WinMain(WahCade):
    """Wah!Cade Main Window"""
    def __init__(self, config_opts):
        """initialise main Wah!Cade window"""
        WahCade.__init__(self)
        #command-line options
        self.config_opts = config_opts
        #default, user config: create if necessary
        #set default config location
        self.config_dir = os.path.expanduser('~/.wahcade')
        if not os.path.exists(self.config_dir):
            #copy ALL config files
            self.copy_user_config('all')
            #now we've copied stuff, quit and tell the user
            print _("Wah!Cade config files created in: [%s]") % (os.path.expanduser('~/.wahcade'))
            print _("Restart after configuring (run wahcade-setup or see the README file).")
            sys.exit(0)
        else:
            #update current config
            self.copy_user_config()
        #build the window
        self.winMain = gtk.Window()
        self.fixd = gtk.Fixed()
        self.imgBackground = gtk.Image()
        self.imgMainLogo = gtk.Image()
        self.lblGameListIndicator = gtk.Label()
        self.lblEmulatorName = gtk.Label()
        self.lblGameSelected = gtk.Label()
        if gst_video_imported:
            self.drwVideo = gst_video.VideoWidget()
        else:
            self.drwVideo = gtk.Image()
        self.imgArtwork1 = gtk.Image()
        self.imgArtwork2 = gtk.Image()
        self.imgArtwork3 = gtk.Image()
        self.imgArtwork4 = gtk.Image()
        self.imgArtwork5 = gtk.Image()
        self.imgArtwork6 = gtk.Image()
        self.imgArtwork7 = gtk.Image()
        self.imgArtwork8 = gtk.Image()
        self.imgArtwork9 = gtk.Image()
        self.imgArtwork10 = gtk.Image()
        self.lblGameDescription = gtk.Label()
        self.lblRomName = gtk.Label()
        self.lblYearManufacturer = gtk.Label()
        self.lblScreenType = gtk.Label()
        self.lblControllerType = gtk.Label()
        self.lblDriverStatus = gtk.Label()
        self.lblCatVer = gtk.Label()
        #create scroll list widget
        self.sclGames = ScrollList()
        #image & label lists
        self._layout_items = [
            (8, self.imgMainLogo),
            (21, self.lblGameListIndicator),
            (34, self.lblEmulatorName),
            (60, self.lblGameSelected),
            (73, self.imgArtwork1),
            (86, self.imgArtwork2),
            (99, self.imgArtwork3),
            (112, self.imgArtwork4),
            (125, self.imgArtwork5),
            (138, self.imgArtwork6),
            (151, self.imgArtwork7),
            (164, self.imgArtwork8),
            (177, self.imgArtwork9),
            (190, self.imgArtwork10),
            (47, self.sclGames),
            (203, self.lblGameDescription),
            (216, self.lblRomName),
            (229, self.lblYearManufacturer),
            (242, self.lblScreenType),
            (255, self.lblControllerType),
            (268, self.lblDriverStatus),
            (281, self.lblCatVer)]
        self._main_images = [
            self.imgArtwork1,
            self.imgArtwork2,
            self.imgArtwork3,
            self.imgArtwork4,
            self.imgArtwork5,
            self.imgArtwork6,
            self.imgArtwork7,
            self.imgArtwork8,
            self.imgArtwork9,
            self.imgArtwork10]
        self.visible_img_list = []
        self.visible_img_paths = []
        self._main_labels = [
            #self.lblGameListIndicator,
            self.lblGameSelected,
            self.lblGameDescription,
            self.lblRomName,
            self.lblYearManufacturer,
            self.lblScreenType,
            self.lblControllerType,
            self.lblDriverStatus,
            self.lblCatVer]
        #add widgets to main window
        self.fixd.add(self.imgBackground)
        self.imgBackground.show()
        self.fixd.show()
        self.winMain.add(self.fixd)
        for line, widget in self._layout_items:
            if widget != self.sclGames:
                self.fixd.add(self.make_evb_widget(widget))
            else:
                self.fixd.add(widget.fixd)
        #video widget
        self.video_playing = False
        self.video_enabled = False
        self.video_timer = None
        self.video_player = None
        #if self.drwVideo is not None:
        self.drwVideo.show()
        self.fixd.add(self.drwVideo)
        #list
        self.sclGames.auto_update = False
        self.sclGames.display_limiters = False
        #set window properties
        self.winMain.set_position(gtk.WIN_POS_NONE)
        self.winMain.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_NORMAL)
        self.winMain.set_title('Wah!Cade')
        #init random engine
        random.seed()
        #build list
        self.lsGames = []
        self.lsGames_len = 0
        #timers
        self.scrsave_time = time.time()
        #self.start_timer('scrsave')
        #get keyboard & mouse events
        self.sclGames.connect('update', self.on_sclGames_changed)
        self.sclGames.connect('mouse-left-click', self.on_sclGames_changed)
        self.sclGames.connect('mouse-right-click', self.on_winMain_key_press)
        self.sclGames.connect('mouse-double-click', self.launch_auto_apps_then_game)
        self.winMain.connect('destroy', self.on_winMain_destroy)
        self.winMain.add_events(
            gtk.gdk.POINTER_MOTION_MASK |
            gtk.gdk.SCROLL_MASK |
            gtk.gdk.BUTTON_RELEASE_MASK |
            gtk.gdk.KEY_PRESS |
            gtk.gdk.KEY_RELEASE)
        key_press_events = [
            'key-press-event', 'key-release-event', 'button-release-event',
            'scroll-event',  'motion-notify-event']
        [self.winMain.connect(kpev, self.on_winMain_key_press) for kpev in key_press_events]
        #key_press_events = ['key-press-event', 'key-release-event']
        #[self.drwVideo.connect(kpev, self.on_winMain_key_press) for kpev in key_press_events]
        self.main_scroll_keys = [
            'UP_1_GAME', 'DOWN_1_GAME', 
            'UP_1_PAGE', 'DOWN_1_PAGE', 
            'UP_1_LETTER', 'DOWN_1_LETTER']
        self.keypress_count = 0 
        #get ini files
        self.wahcade_ini = MameWahIni(os.path.join(self.config_dir, 'wahcade.ini'))
        self.current_emu = self.wahcade_ini.get('current_emulator')
        self.ctrlr_ini = MameWahIni(os.path.join(self.config_dir, 'ctrlr', 'default.ini'), 'ctrlr')
        self.keep_aspect = self.wahcade_ini.getint('keep_image_aspect')
        self.scrsave_delay = self.wahcade_ini.getint('delay')
        #create options window
        self.options = WinOptions(self)
        self.options.winOptions.hide()
        #create message window
        self.message = WinMessage(self)
        self.message.winMessage.hide()
        #create screen saver window
        self.scrsaver = WinScrSaver(self)
        self.scrsaver.winScrSaver.hide()
        self.scrsave_timer = None
        #add options, message & screen saver widgets to _layout_items
        self._layout_items += [
            (301, self.options.lblHeading),
            (314, self.options.sclOptions),
            (327, self.options.lblSettingHeading),
            (340, self.options.lblSettingValue),
            (357, self.message.lblHeading),
            (370, self.message.lblMessage),
            (383, self.message.lblPrompt),
            (396, self.scrsaver.imgArtwork1),
            (409, self.scrsaver.imgArtwork2),
            (422, self.scrsaver.imgArtwork3),
            (435, self.scrsaver.imgArtwork4),
            (448, self.scrsaver.imgArtwork5),
            (461, self.scrsaver.imgArtwork6),
            (474, self.scrsaver.imgArtwork7),
            (487, self.scrsaver.imgArtwork8),
            (500, self.scrsaver.imgArtwork9),
            (513, self.scrsaver.imgArtwork10),
            (526, self.scrsaver.lblGameDescription),
            (539, self.scrsaver.lblMP3Name)]
        #build list of emulators
        self.emu_lists = []
        emu_ini_files = glob.glob(os.path.join(self.config_dir, 'ini', '*[!0-9].ini'))
        for emu_ini in emu_ini_files:
            ini = MameWahIni(emu_ini)
            self.emu_lists.append(
                [ini.get('emulator_title'),
                os.path.splitext(os.path.basename(emu_ini))[0]])
        self.emu_lists.sort()
        #load list
        self.current_list_ini = None
        self.emu_ini = None
        self.layout_file = ''
        self.load_emulator()
        #fullscreen? and hide cursor & video widget
        self.start_timer('scrsave')
        self.winMain.show()
        self.drwVideo.set_property('visible', False)
        self.sclGames.use_mouse = self.ctrlr_ini.getint('mouse')
        self.sclGames.wrap_list = self.wahcade_ini.getint('wrap_list')
        if not self.wahcade_ini.getint('show_cursor'):
            self.hide_mouse_cursor(self.winMain)
        #fullscreen    
        #if not(self.config_opts.fullscreen or self.config_opts.windowed):
        if self.config_opts.fullscreen and not self.config_opts.windowed:    
            self.winMain.fullscreen()
        elif self.config_opts.windowed and not self.config_opts.fullscreen:
            pass
        elif self.wahcade_ini.getint('fullscreen', 0):
            self.winMain.fullscreen()

    def on_winMain_destroy(self, *args):
        """done, quit the application"""
        #stop vid playing if necessary
        self.stop_video()
        #save ini files
        self.wahcade_ini.write()
        self.emu_ini.write()
        self.current_list_ini.write()
        #write fav list
        filters.write_fav_list(
            os.path.join(self.config_dir, 'files', '%s.fav' % (self.current_emu)),
            self.emu_favs_list)
        #exit gtk loop
        gtk.main_quit()

    def on_winMain_key_press(self, widget, event, *args):
        """key pressed - translate to mamewah setup"""
        mw_keys = []
        mw_key = ''
        mw_func = ''
        mw_functions = []
        #reset screen saver time (and turn off if necessary)
        self.scrsave_time = time.time()
        if self.scrsaver.running:
            self.scrsaver.stop_scrsaver()
            self.start_timer('scrsave')
        if event.type == gtk.gdk.MOTION_NOTIFY:
            #mouse moved
            return
        elif event.type == gtk.gdk.BUTTON_RELEASE:
            #mouse click
            mw_keys = ['MOUSE_BUTTON%s' % (event.button - 1)]
        elif event.type == gtk.gdk.SCROLL: 
            #mouse scroll wheel
            if event.direction in [gtk.gdk.SCROLL_UP, gtk.gdk.SCROLL_LEFT]:
                mw_keys = ['MOUSE_SCROLLUP']
            elif event.direction in [gtk.gdk.SCROLL_DOWN, gtk.gdk.SCROLL_RIGHT]:
                mw_keys = ['MOUSE_SCROLLDOWN']
            else:
                return
        elif event.type == gtk.gdk.KEY_PRESS:
            self.keypress_count += 1 
            #keyboard pressed, get gtk keyname
            keyname = gtk.gdk.keyval_name(event.keyval).lower()
            #got something?
            if not mamewah_keys.has_key(keyname):
                return
            #get mamewah keyname
            mw_keys = mamewah_keys[keyname]
            if mw_keys == []:
                return
        elif event.type == gtk.gdk.KEY_RELEASE:
            self.keypress_count = 0
            #keyboard released, update labels, images, etc
            if widget == self.winMain:
                self.sclGames.update()
        #get mamewah function from key
        for mw_key in mw_keys:
            mw_functions = self.ctrlr_ini.reverse_get(mw_key)
            if mw_functions:
                break
        for mw_func in mw_functions:
            #which function?
            if widget == self.winMain:
                #stop vid playing now if the list is likely to change
                #if mw_func in self.main_scroll_keys:
                #    self.stop_video()
                #main form
                if mw_func == 'UP_1_GAME':
                    self.sclGames.scroll((int(self.keypress_count / 20) * -1) - 1) 
                elif mw_func == 'DOWN_1_GAME':
                    self.sclGames.scroll(int(self.keypress_count / 20) + 1) 
                elif mw_func == 'UP_1_PAGE':
                    self.sclGames.scroll(-self.sclGames.num_rows)
                elif mw_func == 'DOWN_1_PAGE':
                    self.sclGames.scroll(+self.sclGames.num_rows)
                elif mw_func == 'UP_1_LETTER':
                    cl = self.lsGames[self.sclGames.get_selected()][0][0]
                    games = [r[0] for r in self.lsGames]
                    games = games[:self.sclGames.get_selected()]
                    games.reverse()
                    i = 0
                    for row in games:
                        i += 1
                        if row[0] != cl:
                            self.sclGames.scroll(-i)
                            break
                elif mw_func == 'DOWN_1_LETTER':
                    cl = self.lsGames[self.sclGames.get_selected()][0][0]
                    games = [r[0] for r in self.lsGames]
                    games = games[self.sclGames.get_selected():]
                    i = -1
                    for row in games:
                        i += 1
                        if row[0] != cl:
                            self.sclGames.scroll(+i)
                            break
                elif mw_func == 'RANDOM_GAME':
                    self.sclGames.set_selected(self.get_random_game_idx())
                elif mw_func == 'FIND_GAME':
                    self.options.set_menu('find')
                    self.options.winOptions.show()
                elif mw_func == 'ADD_GAME_TO_LIST':
                    self.options.set_menu('add_to_list')
                    self.options.winOptions.show()
                elif mw_func == 'REMOVE_GAME_FROM_LIST':
                    self.remove_current_game()
                elif mw_func == 'LAUNCH_GAME':
                    self.launch_auto_apps_then_game()
                elif mw_func == 'LAUNCH_GAME_WITH_OPTIONS1':
                    self.launch_auto_apps_then_game(
                        self.emu_ini.get('alt_commandline_format_1'))
                elif mw_func == 'LAUNCH_GAME_WITH_OPTIONS2':
                    self.launch_auto_apps_then_game(
                        self.emu_ini.get('alt_commandline_format_2'))
                elif mw_func == 'MENU_SHOW':
                    self.options.set_menu('main')
                    self.options.winOptions.show()
                elif mw_func == 'SELECT_EMULATOR':
                    self.options.set_menu('emu_list')
                    self.options.winOptions.show()
                elif mw_func == 'NEXT_EMULATOR':
                    emu_list = [e[1] for e in self.emu_lists]
                    emu_idx = emu_list.index(self.current_emu)
                    if emu_idx < len(emu_list) - 1:
                        emu_idx += 1
                    else:
                        emu_idx = 0
                    self.load_emulator(emu_list[emu_idx])
                elif mw_func == 'PREVIOUS_EMULATOR':
                    emu_list = [e[1] for e in self.emu_lists]
                    emu_idx = emu_list.index(self.current_emu)
                    if emu_idx > 0:
                        emu_idx -= 1
                    else:
                        emu_idx = len(emu_list) - 1
                    self.load_emulator(emu_list[emu_idx])
                elif mw_func == 'SELECT_GAMELIST':
                    self.options.set_menu('game_list')
                    self.options.winOptions.show()
                elif mw_func == 'NEXT_GAMELIST':
                    self.current_list_idx = self.get_next_list_in_cycle(+1)
                    self.load_list()
                elif mw_func == 'PREVIOUS_GAMELIST':
                    self.current_list_idx = self.get_next_list_in_cycle(-1)
                    self.load_list()
                elif mw_func == 'LAUNCH_APP_1':
                    self.external_app_queue = []
                    self.launch_external_application(1)
                elif mw_func == 'LAUNCH_APP_2':
                    self.external_app_queue = []
                    self.launch_external_application(2)
                elif mw_func == 'LAUNCH_APP_3':
                    self.external_app_queue = []
                    self.launch_external_application(3)
                elif mw_func == 'EXIT_TO_WINDOWS':
                    self.on_winMain_destroy()
            elif widget == self.options.winOptions:
                #options form
                if mw_func == 'OP_UP_1_OPTION':
                    self.options.sclOptions.scroll(-1)
                elif mw_func == 'OP_DOWN_1_OPTION':
                    self.options.sclOptions.scroll(+1)
                elif mw_func == 'OP_UP_1_OPTION_PAGE':
                    self.options.sclOptions.scroll(-self.options.sclOptions.num_rows)
                elif mw_func == 'OP_DOWN_1_OPTION_PAGE':
                    self.options.sclOptions.scroll(+self.options.sclOptions.num_rows)
                elif mw_func == 'OP_MENU_SELECT':
                    self.options.menu_selected()
                elif mw_func == 'OP_MENU_HIDE':
                    self.options.winOptions.hide()
                elif mw_func == 'OP_MENU_BACK':
                    if self.options.current_menu == 'main':
                        self.options.winOptions.hide()
                    elif self.options.current_menu in ['emu_list', 'game_list', 'list_options']:
                        self.options.set_menu('main')
                    elif self.options.current_menu == 'add_to_list':
                        self.options.set_menu('list_options')
                    elif self.options.current_menu == 'generate_ftr':
                        if self.current_filter_changed:
                            #generate new filter
                            self.options.set_menu('generate_new_list')
                        else:
                            self.options.set_menu('list_options')
                    elif self.options.current_menu.startswith('ftr:'):
                        self.options.set_menu('generate_ftr')
                    elif self.options.current_menu == 'find':
                        self.options.find_game('back')
            elif widget == self.scrsaver.winScrSaver:
                #screensaver
                if mw_func == 'SS_FIND_N_SELECT_GAME':
                    self.sclGames.set_selected(self.scrsaver.game_idx)
            #history viewer
            if self.histview:
                if widget == self.histview.winHistory:
                    if mw_func == 'UP_1_GAME':
                        self.histview.sclHistory.scroll((int(self.keypress_count / 20) * -1) - 1) 
                    elif mw_func == 'DOWN_1_GAME':
                        self.histview.sclHistory.scroll(int(self.keypress_count / 20) + 1) 
                    elif mw_func in ['UP_1_PAGE', 'UP_1_LETTER']:
                        self.histview.sclHistory.scroll(-self.histview.sclHistory.num_rows)
                    elif mw_func in ['DOWN_1_PAGE', 'DOWN_1_LETTER']:
                        self.histview.sclHistory.scroll(+self.histview.sclHistory.num_rows)
                    elif mw_func in [
                            'EXIT_TO_WINDOWS', 
                            'LAUNCH_APP_%s' % self.histview.app_number, 
                            'LAUNCH_GAME']:
                        self.histview.winHistory.hide()
                        self.auto_launch_external_app()
            #control panel viewer
            if self.cpviewer:
                if widget == self.cpviewer.winCPViewer:
                    if mw_func in [
                            'EXIT_TO_WINDOWS', 
                            'LAUNCH_APP_%s' % self.cpviewer.app_number, 
                            'LAUNCH_GAME']:
                        self.cpviewer.winCPViewer.hide()
                        self.auto_launch_external_app()

    def on_sclGames_changed(self, *args):
        """game selected"""
        #print "on_sclGames_changed: sel=", self.sclGames.get_selected()
        self.stop_video()
        if self.sclGames.get_selected() > self.lsGames_len - 1:
            #blank labels & images
            for img in self.visible_img_list:
                img.set_from_file(None)
            for lbl in self._main_labels:
                lbl.set_text('')
            return
        #set current game in ini file
        self.current_list_ini.set('current_game', self.sclGames.get_selected())
        #get info
        game_info = filters.get_game_dict(self.lsGames[self.sclGames.get_selected()])
        #set layout text items
        self.lblGameDescription.set_text(game_info['game_name'])
        self.lblGameSelected.set_text(_('Game %s of %s selected') % ( 
            self.sclGames.get_selected() + 1, 
            self.lsGames_len))
        if game_info['clone_of'] != '':
            rom_name_desc = _('%s (Clone of %s)') % (game_info['rom_name'], game_info['clone_of'])
        else:
            rom_name_desc = game_info['rom_name']
        self.lblRomName.set_text(rom_name_desc)
        self.lblYearManufacturer.set_text('%s %s' % (game_info['year'], game_info['manufacturer']))
        self.lblScreenType.set_text('%s %s' % (game_info['screen_type'], game_info['display_type']))
        self.lblControllerType.set_text(game_info['controller_type'])
        self.lblDriverStatus.set_text('%s, %s, %s' % (
            game_info['driver_status'],
            game_info['colour_status'],
            game_info['sound_status']))
        self.lblCatVer.set_text(game_info['category'])
        #start video timer
        self.start_timer('video')
        #set layout images
        for i, img in enumerate(self.visible_img_list):
            img_filename = self.get_artwork_image(
                self.visible_img_paths[i],
                self.layout_path,
                game_info,
                self.current_emu,
                (i + 1))
            self.display_scaled_image(img, img_filename, self.keep_aspect)

    def on_scrsave_timer(self):
        """timer event - check to see if we need to start video or screen saver"""
        #need to start screen saver?
        if int(time.time() - self.scrsave_time) >= self.scrsave_delay:
            #yes, stop any vids playing
            self.stop_video()
            if self.emu_ini.get('saver_type') in ['blank_screen', 'slideshow', 'movie', 'launch_scr']:
                #start screen saver
                self.scrsaver.start_scrsaver(self.emu_ini.get('saver_type'))
            else:
                print _("Error: wrong screen saver type: [%s]") % (self.emu_ini.get('saver_type'))
            return False
        #done
        return True

    def on_video_timer(self):
        """timer event - start video playing"""
        #start video playing?
        if self.video_enabled:
            #stop existing vid playing
            self.stop_video()
            #something in the list?
            if self.lsGames_len == 0:
                return False
            #get info
            vid_filename = self.get_video_file(
                self.emu_ini.get('movie_preview_path'),
                filters.get_game_dict(self.lsGames[self.sclGames.get_selected()]))
            if os.path.isfile(vid_filename):
                #resize video vidget
                self.video_playing = True
                img_w, img_h = self.video_artwork_widget.get_size_request()
                xpos = (img_w - self.video_width) / 2
                self.fixd.move(
                    self.drwVideo,
                    self.fixd.child_get_property(self.video_artwork_widget.get_parent(), 'x') + xpos,
                    self.fixd.child_get_property(self.drwVideo, 'y'))
                self.drwVideo.set_size_request(self.video_width, self.video_height)
                self.drwVideo.set_property('visible', True)
                #start video
                self.video_player.play(vid_filename)
        #done
        return False
        
    def stop_video(self):
        """stop and playing video and timer"""
        if self.video_timer:
            gobject.source_remove(self.video_timer)
        if self.video_playing:
            self.video_playing = False
            self.video_player.stop()
            self.drwVideo.set_property('visible', False)
            self.do_events()
        
    def get_next_list_in_cycle(self, direction):
        """return index of next "cycleable" list in current emulator lists"""
        current_idx = self.current_list_idx
        new_idx = current_idx 
        cnt = 0
        #loop round lists until next one with cycle_list set is found
        while cnt < MAX_LISTS:
            cnt += 1
            new_idx += direction
            #wrap tests
            if new_idx < 0:
                new_idx = len(self.game_lists) - 1
            elif new_idx >= len(self.game_lists):
                new_idx = 0
            #does this new list have 'cycle_list' set to True?
            if (self.game_lists[new_idx][2] == 1):
                current_idx = new_idx
                break
        return current_idx
                        
    def launch_auto_apps_then_game(self, game_cmdline_args=''):
        """call any automatically launched external applications 
           then run currently selected game"""
        self.external_app_queue = self.emu_ini.get('auto_launch_apps').split(',')
        self.external_app_queue.reverse() #get it into correct order
        if self.external_app_queue == ['']:
            self.external_app_queue = []
        if len(self.external_app_queue) > 0:
            self.auto_launch_external_app(True, game_cmdline_args)
        else:
            self.launch_game(game_cmdline_args)

    def launch_game(self, cmdline_args=''):
        """run currently selected game"""
        #collect any memory "leaks"
        gc.collect()
        #stop any vids playing
        self.stop_video()
        #get rom name
        rom = self.lsGames[self.sclGames.get_selected()][GL_ROM_NAME]
        #show launch message
        self.message.display_message(
            _('Starting...'),
            '%s: %s' % (rom, self.lsGames[self.sclGames.get_selected()][GL_GAME_NAME]))
        #start timing
        time_game_start = time.time()
        #wait a bit - to let message window display
        while True:
            self.do_events()
            if (time_game_start + 0.2) < time.time():
                break
        #get command line options
        if cmdline_args:
            opts = cmdline_args
        else:
            opts = self.current_list_ini.get(
                option = 'commandline_format',
                default_value = self.emu_ini.get('commandline_format'))
        #minimize?
        minimize_wahcade = ('{minimize}' in opts)
        #replace markers with actual values
        opts = opts.replace('[name]', rom)
        opts = opts.replace('[rompath]', self.emu_ini.get('rom_path'))
        opts = opts.replace('[romext]', self.emu_ini.get('rom_extension'))
        opts = opts.replace('{minimize}', '')
        #set command line
        cmd = '%s %s' % (self.emu_ini.get('emulator_executable'), opts)
        if not self.config_opts.debug and sys.platform != 'win32':
            #write to emulator.log file
            log_filename = os.path.join(self.config_dir, 'emulator.log')
            f = open(log_filename, 'w')
            f.write('********\nCommand from Wah!Cade is:\n  %s\n********\n\n' % cmd)
            f.close()
            #redirect output to log file
            cmd = '%s >> %s 2>&1' % (cmd, log_filename)
        if minimize_wahcade:
            self.winMain.iconify()
            self.do_events()
        #run emulator & wait for it to finish
        p = Popen(cmd, shell=True)
        sts = p.wait()
        #hide message window
        self.message.hide()
        self.scrsave_time = time.time()
        #un-minimize
        if minimize_wahcade:
            self.winMain.present()
            self.do_events()
        #start timers again
        self.start_timer('scrsave')
        self.start_timer('video')
        #stop timing
        time_game_stop = time.time()
        #add to / update favs list
        if not self.emu_favs_list.has_key(rom):
            self.emu_favs_list[rom] = [
                rom,
                self.lsGames[self.sclGames.get_selected()][GL_GAME_NAME],
                0,
                0]
        self.emu_favs_list[rom][FAV_TIMES_PLAYED] += 1
        self.emu_favs_list[rom][FAV_MINS_PLAYED] += int((time_game_stop - time_game_start) / 60)

    def auto_launch_external_app(self, launch_game_after=False, cmdline_args=''):
        """launch next app in list, then launch game"""
        if launch_game_after:
            self.launch_game_after = True
        if len(self.external_app_queue) > 0:
            self.launch_external_application(self.external_app_queue.pop(), True)
        elif self.launch_game_after:
            self.launch_game_after = False
            self.launch_game(cmdline_args)

    def launch_external_application(self, app_number, wait_for_finish=False, game_cmdline_args=''):
        """launch app specified in emu.ini"""
        #get app name
        app_name = self.emu_ini.get('app_%s_executable' % (app_number))
        app_params = self.emu_ini.get('app_%s_commandline_format' % (app_number))
        #pre-defined?
        if app_name == 'wahcade-history-viewer':
            if self.histview:
                #set app number so histview can be closed by same keypress that started it
                self.histview.app_number = app_number
                #display game history
                self.histview.set_history(
                    self.lsGames[self.sclGames.get_selected()][GL_ROM_NAME],
                    self.lsGames[self.sclGames.get_selected()][GL_GAME_NAME])
            else:
                self.auto_launch_external_app(game_cmdline_args)
        elif app_name == 'wahcade-cp-viewer':
            if self.cpviewer:
                #set app number so cpviewer can be closed by same keypress that started it
                self.cpviewer.app_number = app_number
                #display control panel info
                cpvw_rom = self.lsGames[self.sclGames.get_selected()][GL_ROM_NAME]
                #use clone name if necessary
                if self.lsGames[self.sclGames.get_selected()][GL_CLONE_OF] != '':
                    cpvw_rom = self.lsGames[self.sclGames.get_selected()][GL_CLONE_OF]
                self.cpviewer.display_game_details(cpvw_rom)
            else:
                self.auto_launch_external_app(game_cmdline_args)
        else:
            #launch the app
            if os.path.isfile(app_name):
                cmd = '%s %s' % (app_name, app_params)
                p = Popen(cmd, shell=True)
                if wait_for_finish:
                    sts = p.wait()
                    self.auto_launch_external_app()

    def load_emulator(self, emulator_name=None):
        """load emulator"""
        self.launch_game_after = False
        #stop any vids playing
        self.stop_video()
        #load emulator ini file
        if emulator_name:
            self.current_emu = emulator_name
            self.wahcade_ini.set('current_emulator', emulator_name)
        #save current emulator list
        if self.emu_ini:
            self.emu_ini.set('current_list', self.current_list_idx)
            self.emu_ini.write()
        #load new emulator
        self.emu_ini = MameWahIni(os.path.join(self.config_dir, 'ini', '%s.ini' % (self.current_emu)))
        self.lblEmulatorName.set_text(self.emu_ini.get('emulator_title'))
        self.game_lists = []
        self.game_lists_normal = []
        #set xml file
        filters._mameinfo_file = os.path.join(self.emu_ini.get('dat_file'))
        filters._catver_ini = os.path.join(self.emu_ini.get('catver_ini_file'))
        #calc number of lists
        for i in range(MAX_LISTS):
            ini_file = os.path.join(self.config_dir, 'ini', '%s-%s.ini' % (self.current_emu, i))
            if os.path.isfile(ini_file):
                ini = MameWahIni(ini_file)
                self.game_lists.append([ini.get('list_title'), i, ini.getint('cycle_list')])
                if i != 0 and ini.get('list_type') == 'normal':
                    self.game_lists_normal.append([ini.get('list_title'), i, ini.getint('cycle_list')])
            else:
                break
        #load fav list
        fav_name = os.path.join(self.config_dir, 'files', '%s.fav' % (self.current_emu))
        if not os.path.isfile(fav_name):
            #create fav list if it doesn't exist
            f = open(fav_name, 'w')
            f.close()
        self.emu_favs_list = filters.read_fav_list(fav_name)
        #play videos?
        self.check_video_settings()
        #load list
        self.current_list_idx = self.emu_ini.getint('current_list')
        self.list_creation_attempted = False
        self.load_list()
        #history viewer?
        self.histview = None
        if self.emu_ini.get('app_1_executable') == 'wahcade-history-viewer':
            self.histview = WinHistory(self)
            if self.histview.winHistory:
                self.histview.winHistory.hide()
        #cp viewer?
        self.cpviewer = None
        if self.emu_ini.get('app_2_executable') == 'wahcade-cp-viewer':
            self.cpviewer = get_viewer_window(self)
            if self.cpviewer:
                self.cpviewer.winCPViewer.hide()

    def load_list(self):
        """load current list"""
        #load layout for new list
        self.load_layout_file()
        #save last list (if defined)
        if self.current_list_ini:
            self.current_list_ini.write()
        #load new list
        self.current_list_ini = MameWahIni(os.path.join(self.config_dir, 'ini', '%s-%s.ini' % (self.current_emu, self.current_list_idx)))
        self.emu_ini.set('current_list', self.current_list_idx)
        #load list & set current game
        self.lblGameListIndicator.set_text(self.current_list_ini.get('list_title'))
        #has initial lst file been created?
        game_list_file = os.path.join(
            self.config_dir,
            'files',
            '%s-0.lst' % (self.current_emu))
        if not os.path.isfile(game_list_file):
            print _('Please Wait. Creating initial filter...')
            self.message.display_message(_('Please Wait'), _('Creating initial filter...'))
            self.list_creation_attempted = True
            self.do_events()
            self.current_list_idx = 0
            filters.create_initial_filter(
                filters._mameinfo_file,
                os.path.join(
                    self.config_dir,
                    'files',
                    '%s-0.ftr' % (self.current_emu)),
                game_list_file,
                self.emu_ini)
            self.load_list()
            #hide message
            self.message.hide()
        #load the list of games
        self.pop_games_list()
        #load the list filter
        self.current_filter_changed = False
        if self.current_emu == 'mame':
            filter_file = os.path.join(
                self.config_dir,
                'files',
                '%s-%s.ftr' % (self.current_emu, self.current_list_idx))
            if not os.path.isfile(filter_file):
                #filter doesn't exist, so try and use use filter for list 0
                filter_file = os.path.join(
                    self.config_dir,
                    'files',
                    '%s-0.ftr' % (self.current_emu))
            self.current_filter = filters.read_filter(filter_file)
        else:
            self.current_filter = None

    def load_layout_file(self):
        """load layout file"""
        #get layout path
        layout_path = os.path.join(self.config_dir, 'layouts', self.wahcade_ini.get('layout'))
        #layout filename - enuname-?.lay
        layout_file = os.path.join(layout_path, '%s-%s.lay' % (self.current_emu, self.current_list_idx))
        if not os.path.isfile(layout_file):
            #layout filename - enuname.lay
            layout_file = os.path.join(layout_path, '%s.lay' % (self.current_emu))
            if not os.path.isfile(layout_file):
                #layout filename - layout.lay
                layout_file = os.path.join(layout_path, 'layout.lay')
                if not os.path.isfile(layout_file):
                    #print _('Warning: Falling back to default classic640x480 layout')
                    layout_file = os.path.join(self.config_dir, 'layouts', 'classic_640x480', 'layout.lay')
        if layout_file == self.layout_file:
            #layout not changed, but emulator has, so... 
            #...build visible lists for displaying artwork images and exit
            self.rebuild_visible_lists()
            return
        self.layout_file = layout_file
        self.layout_path = layout_path
        #read file & strip any crap
        lines = open(self.layout_file, 'r').readlines()
        lines = [s.strip() for s in lines]
        lines.insert(0, '.')
        #window sizes
        main_width, main_height = int(lines[1].split(';')[0]), int(lines[2])
        opt_width, opt_height = int(lines[294].split(';')[0]), int(lines[295])
        msg_width, msg_height = int(lines[353].split(';')[0]), int(lines[354])
        #main window
        self.winMain.set_size_request(main_width, main_height)
        self.winMain.set_default_size(main_width, main_height)
        bg_col = gtk.gdk.color_parse(self.get_colour(int(lines[3])))
        self.winMain.modify_bg(gtk.STATE_NORMAL, bg_col)
        self.fixd.move(self.imgBackground, 0, 0)
        self.imgBackground.set_size_request(main_width, main_height)
        if os.path.isfile(self.get_path(lines[4])):
            self.imgBackground.set_property('visible', True)
            self.imgBackground.set_from_file(self.get_path(lines[4]))
        else:
            self.imgBackground.set_property('visible', False)
        #set options window
        self.options.winOptions.set_size_request(opt_width, opt_height)
        self.options.winOptions.set_default_size(opt_width, opt_height)
        bg_col = gtk.gdk.color_parse(self.get_colour(int(lines[296])))
        self.options.winOptions.modify_bg(gtk.STATE_NORMAL, bg_col)
        self.options.fixd.move(self.options.imgBackground, 0, 0)
        self.options.imgBackground.set_size_request(opt_width, opt_height)
        if os.path.isfile(self.get_path(lines[297])):
            self.options.imgBackground.set_from_file(self.get_path(lines[297]))
            self.options.imgBackground.set_property('visible', True)
        else:
            self.options.imgBackground.set_property('visible', False)
        #games list highlight colours
        hl_bg_col = gtk.gdk.color_parse(self.get_colour(int(lines[6])))
        hl_fg_col = gtk.gdk.color_parse(self.get_colour(int(lines[7])))
        self.sclGames.modify_highlight_bg(gtk.STATE_NORMAL, hl_bg_col)
        self.sclGames.modify_highlight_fg(gtk.STATE_NORMAL, hl_fg_col)
        #options list highlight colours
        hl_bg_col = gtk.gdk.color_parse(self.get_colour(int(lines[299])))
        hl_fg_col = gtk.gdk.color_parse(self.get_colour(int(lines[300])))
        self.options.sclOptions.modify_highlight_bg(gtk.STATE_NORMAL, hl_bg_col)
        self.options.sclOptions.modify_highlight_fg(gtk.STATE_NORMAL, hl_fg_col)
        #set message window
        self.message.winMessage.set_size_request(msg_width, msg_height)
        self.message.winMessage.set_default_size(msg_width, msg_height)
        bg_col = gtk.gdk.color_parse(self.get_colour(int(lines[355])))
        self.message.winMessage.modify_bg(gtk.STATE_NORMAL, bg_col)
        self.message.fixd.move(self.message.imgBackground, 0, 0)
        self.message.imgBackground.set_size_request(msg_width, msg_height)
        if os.path.isfile(self.get_path(lines[356])):
            self.message.imgBackground.set_from_file(self.get_path(lines[356]))
            self.message.imgBackground.set_property('visible', False)
        else:
            self.message.imgBackground.set_property('visible', False)
        #screen saver window
        self.scrsaver.winScrSaver.set_size_request(main_width, main_height)
        self.scrsaver.winScrSaver.set_default_size(main_width, main_height)
        self.scrsaver.winScrSaver.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse('black'))
        #set all window items
        for offset, widget in self._layout_items:
            #get properties
            d = self.get_layout_item_properties(lines, offset)
            #font
            fd = d['font']
            if d['font-bold']:
                fd += ' Bold'
            fd += ' %s' % (d['font-size'])
            font_desc = pango.FontDescription(fd)
            #text colour
            fg_col = gtk.gdk.color_parse(d['text-col'])
            widget.modify_font(font_desc)
            widget.modify_fg(gtk.STATE_NORMAL, fg_col)
            #background colour & transparency
            bg_col = gtk.gdk.color_parse(d['background-col'])
            parent = widget.get_parent()
            if parent.get_ancestor(gtk.EventBox):
                if d['transparent']:
                    parent.set_visible_window(False)
                else:
                    parent.set_visible_window(True)
                    parent.modify_bg(gtk.STATE_NORMAL, bg_col)
            #alignment
            if d['text-align'] == 2:
                widget.set_property('xalign', 0.5)
            else:
                widget.set_property('xalign', d['text-align'])
            #visible?
            if not d['visible']:
                widget.hide()
                if parent.get_ancestor(gtk.EventBox):
                    parent.hide()
            else:
                widget.show()
                if parent.get_ancestor(gtk.EventBox):
                    parent.show()
            #move & size
            widget.set_size_request(d['width'], d['height'])
            #position video widget
            if self.emu_ini.getint('movie_artwork_no') > 0:
                self.video_artwork_widget = self._main_images[(self.emu_ini.getint('movie_artwork_no') - 1)]
                if widget == self.video_artwork_widget:
                    self.fixd.move(self.drwVideo, d['x'], d['y'])
                    self.drwVideo.set_size_request(d['width'], d['height'])
            #modify widget for lists
            if widget == self.sclGames:
                widget = self.sclGames.fixd
            elif widget == self.options.sclOptions:
                widget = self.options.sclOptions.fixd
            elif parent.get_ancestor(gtk.EventBox):
                widget = widget.get_parent()
            #add to fixed layout on correct window
            if offset < 293:
                #main window
                self.fixd.move(widget, d['x'], d['y'])
            elif offset < 353:
                #options window
                self.options.fixd.move(widget, d['x'], d['y'])
            elif offset < 396:
                #message window
                self.message.fixd.move(widget, d['x'], d['y'])
            else:
                #screen saver window
                self.scrsaver.fixd.move(widget, d['x'], d['y'])
        #other stuff
        self.options.lblHeading.set_text(_('Options'))
        self.options.lblSettingHeading.set_text(_('Current Setting:'))
        self.options.lblSettingValue.set_text('')
        #build visible lists for displaying artwork images
        self.rebuild_visible_lists()
        
    def rebuild_visible_lists(self):
        """get list of visible images & paths"""
        self.visible_img_list = [img for img in self._main_images if img.get_property('visible')]
        self.visible_img_paths = [self.emu_ini.get('artwork_%s_image_path' % (i + 1)) for i, img in enumerate(self.visible_img_list)]

    def pop_games_list(self):
        """populate games list"""
        #which type of list is it?
        if self.current_list_idx == 0 or self.current_list_ini.get('list_type') == 'normal':
            #normal, so sort list
            list_filename = os.path.join(
                self.config_dir,
                'files',
                '%s-%s.lst' % (self.current_emu, self.current_list_idx))
            if os.path.isfile(list_filename):
                self.lsGames, self.lsGames_len = filters.read_filtered_list(list_filename)
                if self.lsGames_len == 0 and self.current_list_idx == 0 and not(self.list_creation_attempted):
                    #try re-creating list
                    print _('Please Wait. Creating initial filter...')
                    self.message.display_message(_('Please Wait'), _('Creating initial filter...'))
                    self.list_creation_attempted = True
                    self.do_events()
                    self.current_list_idx = 0
                    filters.create_initial_filter(
                        filters._mameinfo_file,
                        os.path.join(
                            self.config_dir,
                            'files',
                            '%s-0.ftr' % (self.current_emu)),
                        list_filename,
                        self.emu_ini)
                    self.load_list()
                    #hide message
                    self.message.hide()
                    return
            else:
                self.lsGames = []
                self.lsGames_len = 0
            self.lsGames.sort()
        elif self.current_list_ini.get('list_type') in ['most_played', 'longest_played']:
            #favs type, so choose sort column
            if self.current_list_ini.get('list_type') == 'most_played':
                sort_column = FAV_TIMES_PLAYED
            else:
                sort_column = FAV_MINS_PLAYED
            #use all games to gen list
            list_filename = os.path.join(
                self.config_dir,
                'files',
                '%s-0.lst' % (self.current_emu))
            if os.path.isfile(list_filename):
                self.lsGames, self.lsGames_len = filters.read_filtered_list(list_filename)
            else:
                self.lsGames = []
                self.lsGames_len = 0
            #create list of roms
            flist_roms = [r[GL_ROM_NAME] for r in self.lsGames]
            #order fav dictionary by number of plays
            favs = list(self.emu_favs_list.values())
            favs.sort(key = itemgetter(sort_column), reverse=True)
            #order filtered list by favs
            flist_sorted = []
            for fav in favs:
                try:
                    idx = flist_roms.index(fav[FAV_ROM_NAME])
                    flist_sorted.append(self.lsGames[idx])
                except ValueError:
                    print "%s not in list" % (fav[FAV_ROM_NAME])
            self.lsGames = flist_sorted
        #setup scroll list
        if self.current_list_idx == 0:
            self.sclGames.ls = [l[0] for l in self.lsGames]
        else:
            self.sclGames.ls = [l[0].split('(')[0].strip() for l in self.lsGames]
        #select game in list
        current_game = self.current_list_ini.getint('current_game')
        if current_game >= self.lsGames_len: #len(self.sclGames.ls):
            current_game = 0
        self.sclGames.set_selected(current_game)
        self.on_sclGames_changed()

    def get_random_game_idx(self):
        """pick a random game index number"""
        return random.randint(0, self.lsGames_len - 1)

    def remove_current_game(self):
        """remove currently selected game from the list"""
        item = self.sclGames.ls.pop(self.sclGames.get_selected())
        item = self.lsGames.pop(self.sclGames.get_selected())
        filters.write_filtered_list(
            os.path.join(self.config_dir, 'files', '%s-%s.lst' % (
                self.current_emu, self.current_list_idx)),
            self.lsGames)
        #update displays
        self.sclGames.set_selected(self.sclGames.get_selected() - 1)
        self.sclGames.update()
        
    def check_video_settings(self):
        """enable video playing?"""
        self.video_enabled = True
        #did gst_video module import ok?
        if not gst_video_imported:
            self.video_enabled = False
        #sorry, the video control class (gst_video.py) won't work in windows
        if sys.platform == 'win32':
            self.video_enabled = False
        #check video path exists
        if not os.path.exists(self.emu_ini.get('movie_preview_path')):
            if self.config_opts.debug:
                print _('Error: Movie Preview Path [%s] does not exist' % (self.emu_ini.get('movie_preview_path')))
            self.video_enabled = False
        #check movie artwork
        if not(self.emu_ini.getint('movie_artwork_no') > 0):
            self.video_enabled = False
        #create gstreamer video player
        if self.video_enabled:
            self.video_player = gst_video.GstVideo(self.drwVideo)
            self.video_player.set_volume(self.wahcade_ini.getint('movie_volume'))

    def start_timer(self, timer_type):
        """start given timer"""
        #screen saver
        if timer_type == 'scrsave' and self.scrsave_delay > 0:
            if self.scrsave_timer:
                gobject.source_remove(self.scrsave_timer)
            self.scrsave_timer = gobject.timeout_add(2500, self.on_scrsave_timer)
        #video
        elif timer_type == 'video' and self.video_enabled:
            #stop any playing vids first
            self.stop_video()
            #restart timer
            self.video_timer = gobject.timeout_add(
                self.wahcade_ini.getint('delay_before_movie_preview') * 1000,
                self.on_video_timer)
 
